/*-------------------------------------------------------------------------
# Author: Federico Guerra
# Date: 4 July 2007
# Department of Information Engineering (DEI), University of Padova
#-------------------------------------------------------------------------
#
# Copyright (c) 2007 Regents of the SIGNET lab, University of Padova.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
# 3. Neither the name of the University of Padova (SIGNET lab) nor the 
#    names of its contributors may be used to endorse or promote products 
#    derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED 
# TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR 
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR 
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; 
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR 
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF 
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.*/

/*-------------------------------------------------------------------------
* Description: 
* This module simulates a voice conversation implementing the Stern Model
*------------------------------------------------------------------------- 
* Usage:
* The module works this way: first create a master and a slave module and
* set their parameters; then attach the slave to the master. 
* Only the master module can receive any start or the stop command.
*-------------------------------------------------------------------------
* Example of usage in TCL: 
* (for the complete list of "set" check cnv_defaults.tcl)
* 
* set cnv_example_master   [new Module/CNV]
*
* $cnv_example_master set framerate_ 0.020
* $cnv_example_master set hangover_ 0
* $cnv_example_master set fillin_ 0
* $cnv_example_master set seed_ 222
* 
* set cnv_example_slave   [new Module/CNV]
*
* $cnv_example_slave set framerate_ 0.020
* $cnv_example_slave set hangover_ 0
* $cnv_example_slave set fillin_ 0
* $cnv_example_slave set seed_ 122
*
* $cnv_example_master attach-slave $cnv_example_slave
*
*--------------------------------------------------------------------------- 
* TCL Command list (for a command explanation search this cc file)
*
* $cnv_example_master start 
* $cnv_example_master start_with_VAD
* $cnv_example_master start_with_VAD_hangover
* $cnv_example_master start_with_VAD_fillin
* $cnv_example_master start_with_fillin
* $cnv_example_master start_with_hangover
* $cnv_example_master stop
*-------------------------------------------------------------------------*/

#include <iostream>
#include <cmath>
#include <node-core.h>
#include <ip.h>

#include "cnv-module.h"


extern packet_t PT_CNV;

int hdr_conversation::offset_;

static class ConversationPktClass : public PacketHeaderClass {
public:
  ConversationPktClass() : PacketHeaderClass("PacketHeader/CNV", sizeof(hdr_conversation)) {
    this->bind();
    bind_offset(&hdr_conversation::offset_);
  }
} class_conversation_pkt; 



static class ConversationModuleClass : public TclClass {
public:
  ConversationModuleClass() : TclClass("Module/CNV") {}
  TclObject* create(int, const char*const*) {
    return (new ConversationModule());
  }
} class_module_conversation;



void SendTimer::expire(Event *e)
{
  module_send->transmit();
}

void StateTimer::expire(Event *e)
{ // we are checking if number of pcks to send is an integer
  double frame_remainder = fmod(module_state->statetime_ , module_state->framerate_);
  if (frame_remainder == 0.0) // if pcks is a perfect integer then change state
  {
     module_state->changeState();
  }  
  else { // else we have to send the remaining packet
     module_state->transmit(); 
     module_state->changeState();
  }
}

int ConversationModule::uidcnt_ ;		// unique id of the packet generated

/*Markov rate transitions q_ij initialization*/
double ConversationModule::alpha[18] = {

	0.27853,//alpha 1,2
	0.83305,  //alpha 1,4
	5.4890,  //alpha 1,7
	2.3245,  //alpha 2,1
	2.1572,  //alpha 2,6
	2.1572, //alpha 3,1
	2.3245, //alpha 3,6
	2.2222, //alpha 4,1
	1.0438, //alpha 4,6
	1.0438, //alpha 5,1
	2.2222, //alpha 5,6
	0.27853, //alpha 6,3
	0.83305, //alpha 6,5
	5.4890, //alpha 6,8
	27.62, //alpha 7,1
	0.27853, //alpha 7,2
	0.27853, //alpha 8,3
	27.62 //alpha 8,6
};

/*Markov chain's lambda_ij initialization */
double ConversationModule::lambda[8] = {

lambda[0]= alpha[0] + alpha[1] + alpha[2],  // lambda 1,1
lambda[1]=  alpha[3] + alpha[4], // lambda 2,2
lambda[2]= alpha[5] + alpha[6] ,  // lambda 3,3
lambda[3]= alpha[7] + alpha[8] ,  // lambda 4,4
lambda[4]= alpha[10] + alpha[9] ,  // lambda 5,5
lambda[5]= alpha[11] + alpha[12] + alpha[13], // lambda 6,6
lambda[6]= alpha[14] + alpha[15], // lambda 7,7
lambda[7]= alpha[16] + alpha[17] // lambda 8,8
};

/* Markov Chain's transition probabilities p_ij  
 * given that we are leaving the current state*/
double ConversationModule::prob[18] = {

prob[0]= alpha[0]/(alpha[0] + alpha[1] + alpha[2]),  //prob 1,2
prob[1]= alpha[1]/(alpha[0] + alpha[1] + alpha[2]),  //prob 1,4
prob[2]= alpha[2]/(alpha[0] + alpha[1] + alpha[2]),  //prob 1,7
prob[3]= alpha[3]/(alpha[3] + alpha[4]),  //prob 2,1
prob[4]= alpha[4]/(alpha[3] + alpha[4]),  //prob 2,6
prob[5]= alpha[5]/(alpha[5] + alpha[6]), //prob 3,1
prob[6]= alpha[6]/(alpha[5] + alpha[6]), //prob 3,6
prob[7]= alpha[7]/(alpha[7] + alpha[8]), //prob 4,1
prob[8]= alpha[8]/(alpha[7] + alpha[8]), //prob 4,6
prob[9]= alpha[9]/(alpha[9] + alpha[10]), //prob 5,1
prob[10]= alpha[10]/(alpha[9] + alpha[10]), //prob 5,6
prob[11]= alpha[11]/(alpha[11] + alpha[12] + alpha[13]), //prob 6,3
prob[12]= alpha[12]/(alpha[11] + alpha[12] + alpha[13]), //prob 6,5
prob[13]= alpha[13]/(alpha[11] + alpha[12] + alpha[13]), //prob 6,8
prob[14]= alpha[14]/(alpha[14] + alpha[15]), //prob 7,1
prob[15]= alpha[15]/(alpha[14] + alpha[15]), //prob 7,2
prob[16]= alpha[16]/(alpha[16] + alpha[17]), //prob 8,3
prob[17]= alpha[17]/(alpha[16] + alpha[17]) //prob 8,6
};


ConversationModule::ConversationModule()  
  : sendTmr_(this),
    stateTmr_(this),
    is_slave(false),
    statetime_(0),
    txsn(1),
    hrsn(0),
    pkts_recv(0),
    pkts_ooseq(0),
    pkts_lost(0),
    pkts_last_reset(0),
    rftt(-1),
    lrtime(0),
    sthr(0)
{
  // binding to TCL variables
  bind("framerate_", &framerate_);
  bind("destPort_", &dstPort_);
  bind("destAddr_", &dstAddr_);
  bind("packetSize_", &pktSize_);
  bind("alpha_", &alpha_);
  bind("fillin_", &fillin_);
  bind("hangover_", &hangover_);
  bind("state_", &state_); //starting state
  bind("seed_", &seed_);
  uidcnt_ = 0;
}

ConversationModule::~ConversationModule()
{
}

// TCL command interpreter
int ConversationModule::command(int argc, const char*const* argv)
{
  Tcl& tcl = Tcl::instance();
  if (argc == 3) {
    /*this command is used to attach the slave module so the master module can set slave's state_ 
	 * and statetime_ */
	if (strcmp(argv[1], "attach-slave") == 0) {
      slave_ = (ConversationModule*) TclObject::lookup(argv[2]); //set the pointer to slave module
      slave_->is_slave = true; //set slave flag
      if ((slave_ == 0) || (is_slave == true) ) {//wrong slave attached or the module given is already slave
		  tcl.resultf("Error: the module is already slave or no such slave %s", argv[2]);
	return(TCL_ERROR);
      }
      return(TCL_OK);
    }
    /*this command is used to detach the slave module */
    if (strcmp(argv[1], "detatch-slave") == 0) {
      slave_ = (ConversationModule*) TclObject::lookup(argv[2]); //set the pointer to slave module
      if (slave_->is_slave == true) slave_->is_slave = false; //detach slave and reset slave flag
      if ((slave_ == 0) || (is_slave == true) ) {
		  tcl.resultf("Error: the module is no slave or no such slave %s", argv[2]);
	return(TCL_ERROR);
      }
      return(TCL_OK);
    } 
  }  
  if(argc==2)
    {
      if(strcasecmp(argv[1], "start") == 0)/* TCL command to start the packet generation and transmission
	                                        * without VAD,nor hangover nor fillin ==> CBR */
	{ 
	  /* slave check */
	  if( (is_slave == true) || (slave_ == 0) )
          {
	    tcl.resultf("Error: the module is slave or slave not attached");
            return(TCL_ERROR);
          }
          else 
          {
            hangover_ON = false;
            vad_ON = false;
	    fillin_ON = false;
	    start();
	    return TCL_OK;
          }
	}
      else if(strcasecmp(argv[1], "stop") == 0)	// TCL command to stop the packet generation
	{
	  if( (is_slave == true) || (slave_ == 0) )
          {
	    tcl.resultf("Error: stopping the slave is not allowed or slave not attached");
            return(TCL_ERROR);
          }
          else 
          {
            stop();
	    return TCL_OK;  
          }
 	}
      else if(strcasecmp(argv[1], "getrtt") == 0)	
	{
	  tcl.resultf("%f",GetRTT());
	  return TCL_OK;
	}
      else if(strcasecmp(argv[1], "getftt") == 0)	
	{
	  tcl.resultf("%f",GetFTT());
	  return TCL_OK;
	}
      else if(strcasecmp(argv[1], "getper") == 0)	
	{
	  tcl.resultf("%f",GetPER());
	  return TCL_OK;
	}	
      else if(strcasecmp(argv[1], "getthr") == 0)	
	{
	  tcl.resultf("%f",GetTHR());
	  return TCL_OK;
	}
      else if(strcasecmp(argv[1], "sendPkt") == 0)	
	{
	  sendPkt();
	  return TCL_OK;
	}
      else if(strcasecmp(argv[1], "nextState") == 0)	
	{
	  nextState();
	  return TCL_OK;
	}
      else if(strcasecmp(argv[1], "start_with_hangover") == 0) // TCL command to start with hangover
	{ /* we do a slave check here */
	  if( (is_slave == true) || (slave_ == 0) )
          {
	    tcl.resultf("Error: the module is slave or slave not attached");
            return(TCL_ERROR);
          }
          else 
          {
            hangover_ON = true;
            vad_ON = false;
	    fillin_ON = false;
	    start();
	    return TCL_OK;
          }
	}
     else if(strcasecmp(argv[1], "start_with_VAD") == 0) //TCL command to start with VAD
	{ /* slave check */
	  if( (is_slave == true) || (slave_ == 0) )
          {
	    tcl.resultf("Error: the module is slave or slave not attached");
            return(TCL_ERROR);
          }
          else 
          {
            hangover_ON = false;
            vad_ON = true;
	    fillin_ON = false;
	    start();
	    return TCL_OK;
          }
	}
     else if(strcasecmp(argv[1], "start_with_VAD_hangover") == 0)// TCL command to start with VAD, hangover	
	{ /* slave check */
	  if( (is_slave == true) || (slave_ == 0) )
          {
	    tcl.resultf("Error: the module is slave or slave not attached");
            return(TCL_ERROR);
          }
          else 
          {
            hangover_ON = true;
            vad_ON = true;
	    fillin_ON = false;
	    start();
	    return TCL_OK;
          }
	}
     else if(strcasecmp(argv[1], "start_with_VAD_fillin") == 0)	// TCL command to start with VAD,fillin
	{ /* slave check */
	  if( (is_slave == true) || (slave_ == 0) )
          {
	    tcl.resultf("Error: the module is slave or slave not attached");
            return(TCL_ERROR);
          }
          else 
          {
            hangover_ON = false;
            vad_ON = true;
	    fillin_ON = true;
	    start();
	    return TCL_OK;
          }
	}
     else if(strcasecmp(argv[1], "start_with_fillin") == 0)	//TCL command to start with fillin
	{
	  if( (is_slave == true) || (slave_ == 0) )
          { /* slave check */
	    tcl.resultf("Error: the module is slave or slave not attached");
            return(TCL_ERROR);
          }
          else 
          {
            hangover_ON = false;
            vad_ON = false;
	    fillin_ON = true;
	    start();
	    return TCL_OK;
          }
	}
     else if(strcasecmp(argv[1], "resetStats") == 0)	
	{
	  resetStats();
	  fprintf(stderr,"ConversationModule::command() resetStats %s, pkts_last_reset=%d, hrsn=%d, txsn=%d\n", tag_ , pkts_last_reset, hrsn, txsn);
	  return TCL_OK;
	}
    }

  return Module::command(argc, argv);
}

int ConversationModule::crLayCommand(ClMessage* m)
{
  switch (m->type()) 
    {
    
      //case whatever: 
      //    return 0;
      //    break;
    
    default:
      return Module::crLayCommand(m);    
    }  
}

int ConversationModule::getState() 
{
  return state_;
}

void ConversationModule::M_initPkt(Packet* p)
{
  hdr_cmn* ch = hdr_cmn::access(p);
  ch->uid() = uidcnt_++;
  ch->ptype() = PT_CNV;
  ch->size() = pktSize_;
  
  hdr_ip* iph = hdr_ip::access(p);
  iph->daddr() = dstAddr_;
  iph->dport() = dstPort_;

  hdr_conversation* cnvh = HDR_CNV(p);
  cnvh->sn = txsn;
  txsn++;
  cnvh->ts = Scheduler::instance().clock();
  ch->timestamp() = cnvh->ts;
  
  // how to set the is_voice flag in Master Mode
  
  /* we preferred to split the cases with "bool voice" and "bool silence" to catch 
   * any eventual error instead of using "voice == true" and "voice == false" in the 
   * if braces below */
  
  /*silence condition is triggered by one of these states: 4,5,6,7,8*/
  bool silence = (state_ == 4) || (state_ == 5) || (state_ == 6) || (state_ == 7) || (state_ == 8) ;
  
  /*voice condition is triggered by one of these states: 1,2,3*/
  bool voice = (state_ == 1) || (state_ == 2) || (state_ == 3) ; 
  
  if ( voice == true ) // we are in the "voice" state
    {
    cnvh->is_voice = true;
    hangover_counter = hangover_pcks; /* reset the hangover counter for the
                                       * next transition to "silence" states */
    }
  else if ( (silence == true ) && (hangover_ON == true) && (hangover_counter > 0) && (fillin_ON == false) )
   { // we are in "silence" BUT hangover is ON and it still has pcks to attach
    cnvh->is_voice = true;
    hangover_counter--; // decrease the counter for the next call of M_inintPck()
   }
  else if ( (silence == true ) && (hangover_ON == true) && (hangover_counter <= 0) && (fillin_ON == false) )
   { // we are in "silence", hangover is ON but there are no pcks left to attach
    cnvh->is_voice = false;
   } 
  else if ( (silence == true ) && (hangover_ON == false) && (fillin_ON == false) ) 
  { //we are in "silence" and hangover and fillin are OFF
   cnvh->is_voice = false;
  }
  else if ( (silence == true ) && (hangover_ON == false) && (fillin_ON == true) )
  { // we are in "silence "with fillin ON
   
   /*hack so we can know in advance how many "silence" pcks will be generated*/
   int silence_pcks = (int)ceil(statetime_/framerate_); 
   int fillin_pcks = (int)ceil(fillin_/framerate_);
   
   if (fillin_pcks >= silence_pcks) 
   { // fillin can cover the "silence" period
    cnvh->is_voice = true;
   }
   else
   { // fillin can't cover the "silence" period
    cnvh->is_voice = false;
   }
  }
  else // catch any error here
  {
  printf("ConversationModule(%s)::M_initPkt() error in is_voice initialization \n",is_slave ? "SLAVE" : "MASTER");
  exit(1);
  }

  if (rftt >= 0)
    {
      cnvh->rftt = rftt;
      cnvh->rftt_valid = true;
    }
  else
    {
      cnvh->rftt_valid = false;
    }

}

void ConversationModule::S_initPkt(Packet* p)
{
  hdr_cmn* ch = hdr_cmn::access(p);
  ch->uid() = uidcnt_++;
  ch->ptype() = PT_CNV;
  ch->size() = pktSize_;
  
  hdr_ip* iph = hdr_ip::access(p);
  iph->daddr() = dstAddr_;
  iph->dport() = dstPort_;

  hdr_conversation* cnvh = HDR_CNV(p);
  cnvh->sn = txsn;
  txsn++;
  cnvh->ts = Scheduler::instance().clock();
  ch->timestamp() = cnvh->ts;
  
  // how to set the is_voice flag in Slave Mode
  
  /* we preferred to split the cases with "bool voice" and "bool silence" to catch 
   * any eventual error instead of using "voice == true" and "voice == false" in the 
   * if braces below */
  
  /*silence condition is triggered by one of these states: 1,4,5,7,8 */
  bool silence = (state_ == 1) || (state_ == 4) || (state_ == 5) || (state_ == 7) || (state_ == 8) ;
  
  /*silence condition is triggered by one of these states: 2,3,6 */
  bool voice = (state_ == 2) || (state_ == 3) || (state_ == 6);
  
  if ( voice == true ) // we are in the "voice" state
    {
    cnvh->is_voice = true;
    hangover_counter = hangover_pcks; /* reset the hangover counter for the
                                       * next transition to "silence" states*/
    }
  else if ( (silence == true ) && (hangover_ON == true) && (hangover_counter > 0) && (fillin_ON == false) )
   { // we are in "silence" BUT hangover is ON and still has pck to attach
    cnvh->is_voice = true;
    hangover_counter--; // decrease the counter for the next call of inintPck()
   }
  else if ( (silence == true ) && (hangover_ON == true) && (hangover_counter <= 0) && (fillin_ON == false) )
   { // we are in "silence", hangover is ON but there are no pcks left to attach
    cnvh->is_voice = false;
   } 
  else if ( (silence == true ) && (hangover_ON == false) && (fillin_ON == false) ) 
  { //we are in "silence" and hangover and fillin are OFF
   cnvh->is_voice = false;
  }
  else if ( (silence == true ) && (hangover_ON == false) && (fillin_ON == true) )
  { // we are in "silence "with fillin ON
   int silence_pcks = (int)ceil(statetime_/framerate_); //silence burst
   int fillin_pcks = (int)ceil(fillin_/framerate_); //max gap admitted 
   if (fillin_pcks >= silence_pcks) 
   { // fillin can cover the "silence" period
    cnvh->is_voice = true;
   }
   else
   { // fillin can't cover the "silence" period
    cnvh->is_voice = false;
   }
  }
  else //catch any error here
  {
  printf("ConversationModule(%s)::S_initPkt() error in is_voice initialization \n",is_slave ? "SLAVE" : "MASTER");
  exit(1);
  }

  if (rftt >= 0)
    {
      cnvh->rftt = rftt;
      cnvh->rftt_valid = true;
    }
  else
    {
      cnvh->rftt_valid = false;
    }

}


void ConversationModule::setStateTime() 
{
   double y = generator1.next_double(); // next uniform double bewteen (0,1)
   statetime_ = ((-1)*log(y)/lambda[state_ - 1]) ; //generates the markov chain's state waiting time
   if (debug_ > 0) {
   printf("ConversationModule(%s)::setStateTime() statetime_ of (%d) = (%f)\n",is_slave ? "SLAVE" : "MASTER",state_,statetime_);	
   }
   slave_->statetime_ = statetime_;
   slave_->state_ = state_;

}

void ConversationModule::nextState()
{
  	if (state_ == 1) { // current state

		double y = generator1.next_double(); //RNG for state transition
		
		if (y <= prob[0]) { 
			state_ = 2; // next state
			state_count[1]++;
			double x = generator1.next_double(); // RNG for waiting time calculation
			statetime_ = ((-1)*log(x)/lambda[1]) ;
		}
		else if ((y > prob[0]) && (y <= (prob[0] + prob[1]))) {
			state_ = 4;
			state_count[3]++;
			double x = generator1.next_double();
			statetime_ = ((-1)*log(x)/lambda[3]) ;
		}
		else if (y > (prob[0] + prob[1]) ) {
			state_ = 7;
			state_count[6]++;
			double x = generator1.next_double();
			statetime_ = ((-1)*log(x)/lambda[6]) ;
		}
		else {
			printf("ConversationModule(%s)::nextState() error in transition from state 1 \n",is_slave ? "SLAVE" : "MASTER"); 
			exit(1);
		}		
	}
		
	else if (state_ == 2)  {
	        double y = generator1.next_double();
		
		if (y <= prob[3]) { 
			state_ = 1;
			state_count[0]++;
			double x = generator1.next_double();
			statetime_ = ((-1)*log(x)/lambda[0]) ;
		}
		else if (y > prob[3])  {
			state_ = 6; 
			state_count[5]++;
			double x = generator1.next_double();
			statetime_ = ((-1)*log(x)/lambda[5]) ;
		}
		else {
			printf("ConversationModule(%s)::nextState() error in transition from state 2 \n",is_slave ? "SLAVE" : "MASTER"); 
			exit(1);
		}		
	}
	else if (state_ == 3)  {
		
		double y = generator1.next_double();
		
		if (y <= prob[5]) { 
			state_ = 1;
			state_count[0]++;
			double x = generator1.next_double();
			statetime_ = ((-1)*log(x)/lambda[0]) ;
		}
		else if (y > prob[5]) {
			state_ = 6; 
			state_count[5]++;
			double x = generator1.next_double();
			statetime_ = ((-1)*log(x)/lambda[5]) ;
		}
		else {
			printf("ConversationModule(%s)::nextState() error in transition from state 3 \n",is_slave ? "SLAVE" : "MASTER"); 
			exit(1);
		}		
	}
	else if (state_ == 4)  {
		
		double y = generator1.next_double();
		
		if (y <= prob[7]) { 
			state_ = 1;
			state_count[0]++;
			double x = generator1.next_double();
			statetime_ = 0.2 + ((-1)*log(x)/lambda[0]) ; // 0.2 + exp as stated in stern paper
		}
		else if (y > prob[7])  {
			state_ = 6; 
			state_count[5]++;
			double x = generator1.next_double();
			statetime_ = 0.2 + ((-1)*log(x)/lambda[5]) ; // 0.2 + exp as stated in stern paper
		}
		else {
			printf("ConversationModule(%s)::nextState() error in transition from state 4 \n",is_slave ? "SLAVE" : "MASTER"); 
			exit(1);
		}		
	}
	else if (state_ == 5)  {
		
		double y = generator1.next_double();
		
		if (y <= prob[9]) { 
			state_ = 1;
			state_count[0]++;
			double x = generator1.next_double();
			statetime_ = 0.2 + ((-1)*log(x)/lambda[0]) ; // 0.2 + exp as stated in stern paper
		}
		else if (y > prob[9]) {
			state_ = 6; 
			state_count[5]++;
			double x = generator1.next_double();
			statetime_ = 0.2 + ((-1)*log(x)/lambda[5]) ; // 0.2 + exp as stated in stern paper
		}
		else {
			printf("ConversationModule(%s)::nextState() error in transition from state 5 \n",is_slave ? "SLAVE" : "MASTER"); 
			exit(1);
		}		
	}
	else if (state_ == 6)  {
		double y = generator1.next_double();
		
		if (y <= prob[11]) { 
			state_ = 3;
			state_count[2]++;
			double x = generator1.next_double();
			statetime_ = ((-1)*log(x)/lambda[2]) ;
		}
		else if ((y > prob[11]) && (y <= (prob[11] + prob[12]))) {
			state_ = 5; 
			state_count[4]++;
			double x = generator1.next_double();
			statetime_ = ((-1)*log(x)/lambda[4]) ;
		}
		else if (y > (prob[11] + prob[12])) {
			state_ = 8;
			state_count[7]++;
			double x = generator1.next_double();
			statetime_ = ((-1)*log(x)/lambda[7]) ;
		}
		else {
			printf("ConversationModule(%s)::nextState() error in transition from state 6 \n",is_slave ? "SLAVE" : "MASTER"); 
			exit(1);
		}		
	}
	else if (state_ == 7)  {
		
	double y = generator1.next_double();
		
		if (y <= prob[14]) { 
			state_ = 1;
			state_count[0]++;
			double x = generator1.next_double();
			statetime_ = ((-1)*log(x)/lambda[0]) ;
		}
		else if (y > prob[14]) {
			state_ = 2; 
			state_count[1]++;
			double x = generator1.next_double();
			statetime_ = ((-1)*log(x)/lambda[1]) ;
		}
		else {
			printf("ConversationModule(%s)::nextState() error in transition from state 7 \n",is_slave ? "SLAVE" : "MASTER"); 
			exit(1);
  		}		
	}
	else if (state_ == 8)  {
		
		double y = generator1.next_double();

		if (y <= prob[16]) { 
			state_ = 3;
			state_count[2]++;
			double x = generator1.next_double();
			statetime_ = ((-1)*log(x)/lambda[2]) ;
		}
		else if (y > prob[16]) {
			state_ = 6; 
			state_count[5]++;
			double x = generator1.next_double();
			statetime_ = ((-1)*log(x)/lambda[5]) ;
		}
		else {
			printf("ConversationModule(%s)::nextState() error in transition from state 8 \n",is_slave ? "SLAVE" : "MASTER"); 
			exit(1);
		}		
	}
	else {
		printf("ConversationModule(%s)::nextState() Unknown error \n",is_slave ? "SLAVE" : "MASTER");
		exit(1);
	}
	
	if (statetime_ < framerate_) //We can't allow statetime_ < time resolution
       	{                            
        	statetime_ = framerate_;  // so we force to last at least a single frame
        }
	slave_->statetime_ = statetime_; //set slave statetime_
        slave_->state_ = state_; //set slave state_
   	if (debug_ > 0) {
   	printf("ConversationModule(%s)::nextState() statetime_ of (%d) = (%f)\n",is_slave ? "SLAVE" : "MASTER",state_,statetime_);	
   	printf("ConversationModule(%s)::nextState() frames to send = %d \n",is_slave ? "SLAVE" : "MASTER",(int)ceil(statetime_/framerate_));
	}
}

/* slave initialization with master parameter */
void ConversationModule::S_initialize() 
{
  slave_->vad_ON = vad_ON;
  slave_->hangover_ON = hangover_ON;
  slave_->fillin_ON = fillin_ON;
  slave_->framerate_ = framerate_;
  slave_->seed_ = seed_;
  slave_->fillin_ = fillin_;
  slave_->pktSize_ = pktSize_;
  slave_->hangover_pcks = hangover_pcks;
  slave_->hangover_counter = hangover_counter;
}

void ConversationModule::start()
{ 
  /* we check the given parameters */
  if ((framerate_ <= 0) || (hangover_ < 0) || (fillin_ < 0))
    {
      fprintf(stderr,"Error ConversationModule(%s)::start()  a parameter is <= 0 \n",is_slave ? "SLAVE" : "MASTER");
      exit(1);
    }
  if ( (state_ < 1) || (state_ > 7)) {
     fprintf(stderr,"Error ConversationModule(%s)::start()  state_ is out of bound [1,8] \n",is_slave ? "SLAVE" : "MASTER");
     exit(1);
  }
  /* generators intialization*/
  //  generator1.set_seed(seed_);
  for (int i = 1; i < seed_ ; i++) {
     generator1.reset_next_substream();
     //generator2.reset_next_substream();
  }
  /* hangover initialization */
  hangover_pcks = (int)ceil(hangover_/framerate_); 
  hangover_counter = 0;
  
  /* state_count initialization */
  for (int i = 1; i < 8; i++) {
     state_count[i]= 0;
  }	
  setStateTime(); // we calculate the first waiting time (statetime_)
  S_initialize(); // we initialize the slave module
  sendTmr_.resched(framerate_); //we reschedule the pck timer
  
  /* if VAD is on we also turn on the markov chain: this is intended to preserve the
   * the cbr behaviour if the VAD is off */
  if (vad_ON == true) stateTmr_.resched(statetime_);
}

void ConversationModule::sendPkt()
{
  double delay;
  
  if (fillin_ON == false) { 
     delay = 0;
  }
  else {
     delay = fillin_;
  }
  Packet* p = Packet::alloc();
  
  /*We check it the module that calls this method is Master or Slave*/
  if (is_slave == false) 
  { //master case
  M_initPkt(p);
  }
  else // slave case
  {
  S_initPkt(p);		
  }
  hdr_cmn* ch = hdr_cmn::access(p);
  hdr_conversation* cnvh = HDR_CNV(p);
  if (debug_>0)
    printf("ConversationModule(%s)::sendPkt(), send a pkt (%d)\n",is_slave ? "SLAVE" : "MASTER", ch->uid());
  
  /*VAD SECTION*/
  if ( (vad_ON == true) && (cnvh->is_voice == true) ) // pck is voice => we can send it
  {
  sendDown(p,delay);
  }
  else if ( (vad_ON == true) && (cnvh->is_voice == false))// pck is silence => we discard it
  {
  if (debug_>0)
    {
    printf("ConversationModule(%s)::sendPkt(), dropped pkt by VAD \n",is_slave ? "SLAVE" : "MASTER", ch->uid());
    }
  Packet::free(p);
  }
  else if ( vad_ON == false ) // if VAD is off we send the pck down
  {
  sendDown(p,delay);
  }
  else 
  {
  printf("ConversationModule(%s)::sendPck(), unknown error occured in VAD section \n",is_slave ? "SLAVE" : "MASTER");  
  exit(1);
  }
}

void ConversationModule::changeState() 
{
  sendTmr_.cancel(); // we stop the current pck timer
  nextState(); // we do a a state transition
  sendTmr_.resched(framerate_);// and then reschedule the state timer with new waiting state time  
  stateTmr_.resched(statetime_); // we reschedule the pck timer
  
}

void ConversationModule::transmit()
{
  sendPkt();
  slave_->sendPkt();
  // schedule next trasnmission
  sendTmr_.resched(framerate_);
}

void ConversationModule::stop()
{
  if (debug_ > 0 ) {
     printStateCount();
    }
  sendTmr_.cancel();
  /*if VAD is ON the we stop the state timer*/
  if (vad_ON == true) stateTmr_.cancel();
}

void ConversationModule::recv(Packet* p, Handler* h)
{
  hdr_cmn* ch = hdr_cmn::access(p);
  recv(p);
}

/* same method of Miracle CbrModule.cc */
void ConversationModule::recv(Packet* p)
{
  hdr_cmn* ch = hdr_cmn::access(p);
  if (debug_>0)     
    printf("ConversationModule(%s)::recv(Packet*p,Handler*) pktId %d\n",is_slave ? "SLAVE" : "MASTER", ch->uid());
    
  if (ch->ptype() != PT_CNV)
    {
      drop(p,1,CNV_DROP_REASON_UNKNOWN_TYPE);
      incrPktInvalid();
      return;
    }

  hdr_ip* iph = hdr_ip::access(p);
  if (iph->saddr() != dstAddr_)
    {
      drop(p,1,CNV_DROP_REASON_WRONG_SADDR);
      incrPktInvalid();
      return;
    }
  if (iph->sport() != dstPort_)
    {
      drop(p,1,CNV_DROP_REASON_WRONG_SPORT);
      incrPktInvalid();
      return;
    }

  hdr_conversation* cnvh = HDR_CNV(p);

  int esn = hrsn + 1; // expected sn
    
  if (cnvh->sn < esn)
    { // packet is out of sequence and is to be discarded
      incrPktOoseq();
      if (debug_ > 0)
	printf("ConversationModule::recv() Pkt out of sequence! cnvh->sn=%d\thrsn=%d\tesn=%d\n",cnvh->sn,hrsn,esn);
      drop(p,1,CNV_DROP_REASON_OUT_OF_SEQUENCE);
      return;
    }
  
  /* a new packet has been received */
  incrPktRecv();
  hrsn = cnvh->sn;  

  if (cnvh->sn > esn)
    { // packet losses are observed
      incrPktLost(cnvh->sn - (esn));     
    }
  
  /* Smoothed throughput calculation */
  double dt = std::max((Scheduler::instance().clock() - lrtime), 1e-6);
  double thr = ch->size() / dt;
  updateThroughput(thr);
  lrtime = Scheduler::instance().clock(); 
    
  /* Forward and Round Trip Time calculation */
  rftt = Scheduler::instance().clock() - cnvh->ts;
  if(cnvh->rftt_valid) 
    {
      updateFTT(cnvh->rftt);
      double rtt = rftt + cnvh->rftt;
      updateRTT(rtt);
    }

  //drop(p,5,CBR_DROP_REASON_RECEIVED);   
  Packet::free(p);

  if (pkts_lost + pkts_recv + pkts_last_reset != hrsn)     
    fprintf(stderr,"ERROR ConversationModule::recv() pkts_lost=%d  pkts_recv=%d  hrsn=%d\n", pkts_lost, pkts_recv, hrsn);

  //assert(pkts_lost + pkts_recv == hrsn);
  
}

/* same methods of Miracle CbrModule.cc */
double ConversationModule::GetRTT() 
{
  return srtt;
}

double ConversationModule::GetFTT() 
{
  return sftt;
}

double ConversationModule::GetPER() 
{
  if ((pkts_recv + pkts_lost)> 0)
    return ((double) pkts_lost / (double)(pkts_recv + pkts_lost));
  else 
    return 0;
}

double ConversationModule::GetTHR() 
{
  return sthr;
}

void ConversationModule::updateRTT(double rtt)
{
      srtt = alpha_ * srtt + (1-alpha_) * rtt;
}

void ConversationModule::updateFTT(double ftt)
{
  sftt = alpha_ * sftt + (1-alpha_) * ftt;
}

void ConversationModule::updateThroughput(double thr)
{
  sthr = alpha_ * sthr + (1-alpha_) * thr;
}

void ConversationModule::incrPktLost(int npkts)
{
  pkts_lost += npkts;
}

void ConversationModule::incrPktRecv()
{
  pkts_recv++;
}

void ConversationModule::incrPktOoseq()
{
  pkts_ooseq++;
}

void ConversationModule::incrPktInvalid()
{
  pkts_invalid++;
}

void ConversationModule::resetStats()
{
  pkts_last_reset += pkts_lost + pkts_recv;
  pkts_recv = 0;
  pkts_ooseq = 0;
  pkts_lost = 0;
  srtt = 0;
  sftt = 0;
  sthr = 0;
  //txsn = 1;
  //hrsn=0;
  rftt = -1;
}

void ConversationModule::printStateCount() 
{
  for (int i = 0; i < 8; i++) {
    
     printf("ConversationModule(%s)::Frequency of state %d is: %d \n",is_slave ? "SLAVE" : "MASTER",i,state_count[i]);
  }


}  

