/*-------------------------------------------------------------------------
# Author: Federico Guerra
# Date: 4 July 2007
# Department of Information Engineering (DEI), University of Padova
#-------------------------------------------------------------------------
#
# Copyright (c) 2007 Regents of the SIGNET lab, University of Padova.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
# 3. Neither the name of the University of Padova (SIGNET lab) nor the 
#    names of its contributors may be used to endorse or promote products 
#    derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED 
# TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR 
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR 
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; 
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR 
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF 
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.*/

#ifndef CONVERSATION_MODULE
#define CONVERSATION_MODULE

#include <module.h>
#include <rng.h>

/* same definitions of Miracle CbrModule.h */
#define CNV_DROP_REASON_UNKNOWN_TYPE "UKT"
#define CNV_DROP_REASON_RECEIVED "RCV"
#define CNV_DROP_REASON_OUT_OF_SEQUENCE "OOS"
#define CNV_DROP_REASON_WRONG_SADDR "WSA"
#define CNV_DROP_REASON_WRONG_SPORT "WSP"

#define HDR_CNV(P)      (hdr_conversation::access(P))

extern packet_t PT_CNV;

typedef struct hdr_conversation {

  double ts;  /**< packet timestamp, i.e., its generation time) */
  int sn;     /**< sequence number of this packet */  
  bool rftt_valid; /**< whether the rftt value is valid (a packet must
		           * have been received in the opposite direction) */
  
  bool is_voice; // to tag if its type is voice or silence
 
  double rftt;    /**< Forward Trip Time seen in the opposite direction
		* (opposite with respect to the direction of this
		* packet). This is used to calculate Round Trip Time 
		*/
  
/*same instructions of Miracle CbrModule.h */
  static int offset_;
  inline static int& offset() { return offset_; }
  inline static struct hdr_conversation* access(const Packet* p) {
    return (struct hdr_conversation*)p->access(offset_);
  }
  
} hdr_conversation;


class ConversationModule;


/**
 * Timer used by ConversationModule to schedule packet transmission
 */
class SendTimer : public TimerHandler
{
  
 public:
  SendTimer(ConversationModule *m) : TimerHandler() { module_send = m; }
	
 protected:
  virtual void expire(Event *e);
  ConversationModule* module_send;

};

/**
 * Timer used by ConversationModule to schedule Markov Chain's state transition
 */
class StateTimer : public TimerHandler
{
  
 public:
  StateTimer(ConversationModule *m) : TimerHandler() { module_state = m; }
	
 protected:
  virtual void expire(Event *e);
  ConversationModule* module_state;
};


/**
 * Implements a Conversation sender/receiver model based on Stern paper 
 */
class ConversationModule : public Module
{
  friend class SendTimer;
  friend class StateTimer;
  
public:

  ConversationModule();
  ~ConversationModule();		
		
  virtual void recv(Packet*);
  virtual void recv(Packet* p, Handler* h);
  virtual int command(int argc, const char*const* argv);
  virtual int crLayCommand(ClMessage* m);
  
  /*same methods of Miracle CbrModule.h */
  virtual double GetRTT();
  virtual double GetFTT();
  virtual double GetPER();  
  virtual double GetTHR(); 

  virtual void resetStats(); 
  virtual void printStateCount();

protected:
  
  ConversationModule* slave_; // Pointer to the slave module

  static double alpha[18]; // transition rates q_ij of the Markov chain
  static double lambda[8]; // exponential parameter of each Markov chain's state
  int state_count[8]; // Markov Chain's states frequency counter
  
  static double prob[18]; /* Markov Chain's transition probabilities p_ij given 
                           * given that we are leaving the current state*/

  double statetime_; /**< state transition timer */
  int state_;   /**<  internal state={0,1}  */
  static int uidcnt_;  /**< simulation-unique packet ID */
  
  nsaddr_t dstPort_;
  nsaddr_t dstAddr_;
  
  int seed_; // seed for the internal RNG generators
  double hangover_; // hangover time in secs
  double fillin_; // fillin time in secs
  bool vad_ON;      //flag for internal case switching
  bool hangover_ON; //flag for internal case switching
  bool fillin_ON; //flag for internal case switching
  
  bool is_slave; // flag to tag if the current module is master or slave
  
  int hangover_pcks; // number of total hangover pcks
  int hangover_counter; //counter for processed hangover pcks
  
  RNG generator1; // random generator for the Markov chain
  //RNG generator2; // random generator for the Markov chain
  
  /**
   * calls M_initPck or S_initPck depending if the module is Master or 
   * slave, the sends the pck down after VAD calculations
   */ 
  virtual void sendPkt();
  
  /**
   * Call sendPkt and schedule a new transmission after period_ seconds
   */
  virtual void transmit();
  
  /** 
  * calls nextState and schedules a new transition after
  * a statetime_ seconds
  */
  virtual void changeState(); 
  
  /** 
  * does a transition in the Markov chain
  */
  virtual void nextState();
  
  /*
  * calculates the waiting time for the current state_
  */
  virtual void setStateTime();
 
  /*
  * returns current state_
  */
  virtual int getState();
  
  /*
  * Initializes the slave module
  */
  virtual void S_initialize();

  
  /** 
   * initialize a packet before transmission
   * 
   * @param p packet to be initialized
   */	
  virtual void M_initPkt(Packet* p);
  virtual void S_initPkt(Packet* p);
  
  /** 
   * start sending packets
   */
  virtual void start();
  
  /** 
   * stop sending packets
   */
  virtual void stop();
  
  /*same methods of Miracle CbrModule.h*/
  virtual void updateRTT(double rtt);
  virtual void updateFTT(double ftt);
  virtual void updateThroughput(double thr);

  virtual void incrPktLost(int npkts);
  virtual void incrPktRecv();
  virtual void incrPktOoseq();
  virtual void incrPktInvalid();

  double alpha_;      /**< smoothing factor used for statistics using
                          first order IIR filter */
	      
 
  
  int txsn;          /**< sequence number of next new packet to be transmitted */
  int hrsn;          /**< highest received sequence number */

  int pkts_recv;     /**< total number of received packets. Packet out
			of sequence are not counted here. */
  int pkts_ooseq;    /**< total number of packets received out of
			sequence  */
  int pkts_lost;     /**< total number of lost packets, including
			packets received out of sequence   */

  int pkts_invalid;  /**< total number of invalid packets received */

  int pkts_last_reset; /**< for error checking after stats are
			  reset. Set to pkts_lost+pkts_recv each time
			  resetStats is called. */ 

  double rftt;       /**< Forward Trip Time seen for last received
			packet */ 

  double srtt;       /**< Smoothed Round Trip Time, calculated as for
			TCP */

  double sftt;       /**< Smoothed Forward Trip Time, calculated as srtt */

  double lrtime;       /**< Time of last packet reception */
  double sthr;       /**< Smoothed throughput calculation */

  double framerate_;      /**< period between two consecutive packet transmissions */
  
  int pktSize_;        /**< packet size */
  
  SendTimer sendTmr_;  /**< timer which schedules packet transmissions */
  
  StateTimer stateTmr_ ; /** < timer which schedules Markov chain state-transitions   */
		
};




#endif /* CONVERSATION_MODULE */
 
